defmodule ClickHouse.Config do
  @moduledoc false

  ################################
  # Public API
  ################################

  @doc """
  Starts a new configuration table.
  """
  @spec new(ClickHouse.client(), keyword()) :: :ok
  def new(client, opts) do
    table = table_name(client)
    :ets.new(table, [:named_table, read_concurrency: true])

    init(table, :interface, opts)
    init(table, :formats, opts)

    :ok
  end

  @doc """
  Returns the network interface of the given ClickHouse client.
  """
  @spec interface(ClickHouse.client()) :: {:error, :not_found} | {:ok, ClickHouse.Interface.t()}
  def interface(client) do
    fetch(client, :interface)
  end

  @doc """
  Returns a format module for the format name of the given ClickHouse client.
  """
  @spec format(ClickHouse.client(), ClickHouse.Format.name()) ::
          {:error, :not_found} | {:ok, ClickHouse.Format.t()}
  def format(client, name) do
    fetch(client, {:format, name})
  end

  ################################
  # Private API
  ################################

  defp fetch(client, key) do
    case :ets.lookup(table_name(client), key) do
      [{^key, value}] -> {:ok, value}
      _ -> {:error, :not_found}
    end
  rescue
    ArgumentError -> {:error, :not_found}
  end

  defp table_name(client) do
    :"#{__MODULE__}.#{client}"
  end

  defp init(table, :interface, opts) do
    interface = Keyword.fetch!(opts, :interface)
    :ets.insert(table, {:interface, interface})
  end

  defp init(table, :formats, opts) do
    formats = Keyword.fetch!(opts, :formats)

    for format <- formats do
      for name <- format.names() do
        :ets.insert(table, {{:format, name}, format})
      end
    end
  end
end
