defmodule ClickHouse.Query do
  @moduledoc """
  A ClickHouse query.
  """

  alias ClickHouse.{Config, DataType}

  @values_regex ~r/\bVALUES|values\b/i
  @format_regex ~r/(?<= (?:FORMAT|format) )(\w+)/
  @replace_regex ~r/\$\d+/

  @derive {Inspect, only: [:statement, :params]}
  @enforce_keys [:client, :statement, :params, :opts]
  defstruct [
    :client,
    :interface,
    :format,
    :command,
    :statement,
    :encoded,
    :params,
    :params_count,
    :opts
  ]

  @typedoc """
  The type of command being exectued in the query.
  """
  @type command :: :create | :select | :insert | :alter | :unknown

  @typedoc """
  A ClickHouse query
  """
  @type t :: %__MODULE__{
          client: ClickHouse.client(),
          interface: ClickHouse.Interface.t() | nil,
          format: ClickHouse.Format.t() | nil,
          command: command(),
          statement: ClickHouse.statement(),
          encoded: iodata() | nil,
          params: ClickHouse.params(),
          params_count: non_neg_integer()
        }

  ################################
  # Public API
  ################################

  @doc """
  Creates a new ClickHouse query.
  """
  @spec new(ClickHouse.client(), ClickHouse.statement(), ClickHouse.params(), keyword()) ::
          ClickHouse.Query.t()
  def new(client, statement, params \\ [], opts \\ []) do
    query = %__MODULE__{
      client: client,
      statement: statement,
      params: params,
      opts: opts
    }

    query
    |> put_interface()
    |> strip_newlines()
    |> parse_command()
    |> parse_format()
    |> parse_params_count()
  end

  @doc """
  Encodes a ClickHouse query.
  """
  @spec encode(ClickHouse.Query.t()) :: ClickHouse.Query.t()
  def encode(%__MODULE__{command: :insert, params_count: count}) when count > 1 do
    raise ArgumentError, "Insert statement can only support a single parameter."
  end

  def encode(%__MODULE__{command: :insert, params_count: 1, params: []}) do
    raise ArgumentError, "Insert statement cannot reference empty params."
  end

  def encode(%__MODULE__{command: :insert, params_count: 0, params: [_ | _]}) do
    raise ArgumentError, "Insert statement with params must include a '$1'."
  end

  def encode(%__MODULE__{command: :insert, format: nil} = query), do: query

  def encode(%__MODULE__{command: :insert} = query) do
    statement = String.trim_trailing(query.statement, "$1")
    params = query.format.encode(query.params)
    %{query | encoded: [statement, params]}
  end

  def encode(%__MODULE__{command: command} = query)
      when command in [:select, :create, :alter, :unknown] do
    statement = :binary.split(query.statement, " ", [:global])
    encoded = do_encode(statement, query.params)
    %{query | encoded: encoded}
  end

  ################################
  # Private API
  ################################

  defp put_interface(query) do
    case Config.interface(query.client) do
      {:ok, interface} -> %{query | interface: interface}
      _ -> query
    end
  end

  defp strip_newlines(query) do
    statement =
      query.statement
      |> String.replace("\n", " ")
      |> String.trim_leading()
      |> String.trim_trailing()

    %{query | statement: statement}
  end

  defp parse_format(%{command: :insert} = query) do
    if Regex.match?(@values_regex, query.statement) do
      %{query | format: ClickHouse.Format.Values}
    else
      do_parse_format(query)
    end
  end

  defp parse_format(query), do: do_parse_format(query)

  defp do_parse_format(query) do
    case Regex.run(@format_regex, query.statement, capture: :first) do
      [format] -> get_format(query, format)
      _ -> query
    end
  end

  defp get_format(query, format) do
    case Config.format(query.client, format) do
      {:ok, format} -> %{query | format: format}
      _ -> query
    end
  end

  defp parse_command(query) do
    command = do_parse_command(query.statement)
    %{query | command: command}
  end

  defp do_parse_command("SELECT" <> _), do: :select
  defp do_parse_command("select" <> _), do: :select
  defp do_parse_command("INSERT" <> _), do: :insert
  defp do_parse_command("insert" <> _), do: :insert
  defp do_parse_command("CREATE" <> _), do: :create
  defp do_parse_command("create" <> _), do: :create
  defp do_parse_command("ALTER" <> _), do: :alter
  defp do_parse_command("alter" <> _), do: :alter
  defp do_parse_command(_), do: :unknown

  defp parse_params_count(query) do
    params_count = @replace_regex |> Regex.scan(query.statement) |> Enum.count()
    %{query | params_count: params_count}
  end

  defp do_encode(parts, params, acc \\ [])

  defp do_encode([], _, acc), do: acc

  defp do_encode(["$" <> part], params, acc) do
    [acc | [do_encode_part(part, params)]]
  end

  defp do_encode(["$" <> part | parts], params, acc) do
    do_encode(parts, params, [acc | [do_encode_part(part, params), " "]])
  end

  defp do_encode([part], _params, acc) do
    [acc | [part]]
  end

  defp do_encode([part | parts], params, acc) do
    do_encode(parts, params, [acc | [part, " "]])
  end

  defp do_encode_part(part, params) when is_binary(part) do
    part = String.to_integer(part)
    do_encode_part(part, params)
  end

  defp do_encode_part(part, params) when part > 0 do
    data = Enum.fetch!(params, part - 1)
    DataType.encode(data)
  end
end
